/*
 * memmap/memmap.c --
 *
 *	This program demonstrates how to allocate memory using
 *      anonymous memory mappings. It also shows how information
 *	about memory mappings can be found on Linux systems.
 */

#define _POSIX_C_SOURCE 200809L
#define _DEFAULT_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/mman.h>

static int pflag = 0, lflag = 0, vflag = 0;
static const char *progname = "memmap";

#ifndef MAP_ANONYMOUS
#define MAP_ANONYMOUS MAP_ANON
#endif

/*
 * Print the memory map of the running process as this allows us to
 * see how the program changes the memory mappings.
 */

static void
pmap(const char *comment)
{
    char cmd[1234];
  
    if (vflag && comment) {
	printf("\n*** %s ***\n", comment);
	fflush(stdout);
    }
    
    if (pflag || lflag) {
        snprintf(cmd, sizeof(cmd), "pmap %s %d", lflag ? "-x" : "", getpid());
        system(cmd);
    }
}

int main(int argc, char *argv[])
{
    int c;
    char *p;
    long pagesize, length;

    while ((c = getopt(argc, argv, "lpvh")) >= 0) {
	switch (c) {
	case 'l':
	    lflag = 1;
	    break;
	case 'p':
	    pflag = 1;
	    break;
	case 'v':
	    vflag = 1;
	    break;
	case 'h':
	    printf("Usage: %s [-l] [-p] [-v]\n",
		   progname);
	    return EXIT_SUCCESS;
	}
    }

    pagesize = sysconf(_SC_PAGE_SIZE);
    if (pagesize == -1) {
	perror("sysconf");
	return EXIT_FAILURE;
    }

    length = 10 * pagesize;			/* 10 pages - why 10? */
    pmap("after main() is called");
    
    p = mmap(NULL, length, PROT_READ | PROT_WRITE,
	     MAP_ANONYMOUS | MAP_PRIVATE, -1, 0);
    if (p == MAP_FAILED) {
	perror("mmap");
	return EXIT_FAILURE;
    }
    pmap("after calling mmap()");
	
    if (mlock(p, length) == -1) {
	perror("mlock");
	return EXIT_FAILURE;
    }
    pmap("after calling mlock()");
    
    p[0] = 0xff;			/* dirty the first page */
    pmap("after writing p[0]");
    
    p[length-1] = 0xff;			/* dirty the last page */
    pmap("after writing p[length-1]");
    
    if (munmap(p, length) == -1) {
	perror("munmap");
	return EXIT_FAILURE;
    }
    pmap("after calling munmap()");
    
    return EXIT_SUCCESS;
}
